/**
 * @file application.cpp
 * @brief Application class
 *
 * @author Eric Butler (edbutler)
 */

#include "application/application.hpp"

#include "application/opengl.hpp"

#include <SDL/SDL_opengl.h>
#include <SDL/SDL.h>

#include <cassert>
#include <cstdlib>
#include <iostream>
#include <map>

namespace _462 {


struct WindowState
{
    // whether we are currently running
    bool running;
} state;



// header function definitions


Window::Window() { }

Window::~Window() { }

static bool initialize_window( int width, int height, const char* title )
{
	SDL_WM_SetCaption( title, title );

	unsigned int bits_per_pixel = SDL_GetVideoInfo()->vfmt->BitsPerPixel;
	unsigned int flags = SDL_OPENGL;

	SDL_GL_SetAttribute( SDL_GL_BUFFER_SIZE, bits_per_pixel );
	SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );

	if ( SDL_SetVideoMode( width, height, bits_per_pixel, flags ) == 0 ) {
        std::cout << "Error initializing SDL surface: " << SDL_GetError() << ", aborting initialization." << std::endl;
        return false;
	}

    return true;
}

static bool initialize_gui()
{
    return true;
}

static void process_events( Application* app, Window* window )
{
    assert( app && window );

    SDL_Event event;
    while ( SDL_PollEvent( &event ) ) {
        switch ( event.type )
        {
        case SDL_QUIT:
            state.running = false;
            break;

        case SDL_KEYDOWN:
            if ( event.key.keysym.sym == SDLK_ESCAPE ) {
                state.running = false;
            }
            break;

        default:
            break;
        }
        app->handle_event( window, event );
    }
}

static void main_loop( Application* app, Window* window, real_t fps )
{
    assert( app && window );
    assert( fps > 0 );

    // maximum number of frame skips before we force a render
    static const int MAX_FRAME_SKIPS = 5;
    // the maximum possible lag time, lag is capped here
    static const int MAX_LAG_TIME = 500;

    state.running = true;
    int period = static_cast< int >( 1000.0 / fps ); // frame period in milliseconds

    // how far we are behind the target framerate
    int lag_time = 0;
    // the number of updates since the last render
    int num_frame_skips = 0;

    while ( state.running ) {
        int start_time = SDL_GetTicks();

        // process events
        process_events( app, window );

        // update
        app->update( window, fps );

        // render, only if we are not too far behind, or if we've skipped too many frames
        if ( lag_time < period || num_frame_skips >= MAX_FRAME_SKIPS ) {
            app->render( window );
            glFlush();
            SDL_GL_SwapBuffers();
            num_frame_skips = 0;
        } else {
            num_frame_skips++;
        }

        // compute sleep time until end of period, may be negative
		int end_time = SDL_GetTicks();
		int time_diff = end_time - start_time;
		int sleep_time = period - time_diff - lag_time;

        // always yield at least once to keep system responsive
        SDL_Delay( 1 );

        if ( sleep_time > 0 ) {
            // sleep until end of period
            while ( int(SDL_GetTicks()) < end_time + sleep_time ) {
                SDL_Delay( 1 );
            }
            // if we oversleep, add to lag. note this must be nonnegative
            // because of the while loop above
            lag_time += SDL_GetTicks() - end_time + sleep_time;
        } else {
            // otherwise mark how much we're falling behind
            // sleep time is negative here, so it's like adding the time we went over
            lag_time -= sleep_time;
        }

        // cap the lag time to avoid lots of skips
        if ( lag_time > MAX_LAG_TIME ) {
            lag_time = MAX_LAG_TIME;
        }
    }
}

int Window::start_application( Application* app, int width, int height, real_t fps, const char* title )
{
    assert( app );
    // only create an app the first time this function is called
    if ( SDL_WasInit( 0 ) )
        return -2;

    // initialize the application state
    Window* window = new Window();

    // init SDL
	if ( SDL_Init( SDL_INIT_VIDEO | SDL_INIT_TIMER ) == -1 ) {
        std::cout << "Error initializing SDL" << std::endl;
        return -1;
	}

    if ( !initialize_window( width, height, title ) )
        goto FAIL;

    if ( !initialize_gui() )
        goto FAIL;


    // initialize the application
    if ( !app->initialize( window ) )
        goto FAIL;


    // add default menu items
    //state->window->add_menu_item( "Save Screenshot", screenshot_callback );
    //state->window->add_menu_item( "Quit", _462::terminate_application );
    //

    // run main loop
    main_loop( app, window, fps );

    // clean up and exit
    app->destroy( window );
    return 0;


  FAIL:
    std::cerr << "Failed to start applcation, aborting\n";
    return -1;
}

void Window::get_dimension( int* width, int* height )
{
    assert( width && height );
    SDL_Surface* surface = SDL_GetVideoSurface();
    assert( surface );
    *width = surface->w;
    *height = surface->h;
}

void Window::terminate_application()
{
    state.running = false;
}

} /* _462 */

