/**
 * @file application.hpp
 * @brief Application class
 *
 * @author Eric Butler (edbutler)
 */

#ifndef _462_APPLICATION_APPLICATION_HPP_
#define _462_APPLICATION_APPLICATION_HPP_

#include "math/math.hpp"

#include <SDL/SDL_events.h>

namespace _462 {

class Window;

class Application
{
public:

    virtual ~Application() { }

    /**
     * Initializes the application.
     * @return true on successful initialization, false otherwise.
     * @remark The opengl context is valid by this point.
     */
	virtual bool initialize( Window* window ) = 0;

    /**
     * Cleans up any resources used by the application.
     */
	virtual void destroy( Window* window ) = 0;

    /**
     * Updates the application for the current frame.
     * @param delta_time The length of the time step in seconds.
     */
	virtual void update( Window* window, real_t delta_time ) = 0;

    /**
     * Renders the application for the current frame.
     */
	virtual void render( Window* window ) = 0;

    /**
     * Handle an event.
     */
    virtual void handle_event( Window* window, const SDL_Event& event ) = 0;

};

class Window
{
public:

    /**
     * Starts the application and returns a window object.
     * Does not return. Creates the window instance.
     */
    static int start_application(
        Application* app,
        int width, int height, real_t fps,
        const char* title );

    /**
     * Query the current dimension of the window.
     */
    void get_dimension( int* width, int* height );

    /**
     * Terminates the application before the next update loop.
     */
    void terminate_application();

private:

    Window();

    ~Window();

};

} /* _462 */

#endif /* _462_APPLICATION_APPLICATION_HPP_ */

