/**
 * @file camera_roam.cpp
 * @brief CameraRoamControl class
 *
 * @author Zeyang Li (zeyangl)
 * @description   this file implements a free roam camera
 */
 

#include "application/camera_roam.hpp"
#include "math/vector.hpp"

namespace _462 {

void CameraRoamControl::on_key_press(const SDL_Event& event)
{
    unsigned char key = event.key.keysym.sym;
    if(event.type == SDL_KEYDOWN || event.type == SDL_KEYUP)
    {
        switch(key) 
        {
        case SDLK_w:
            move.w = !move.w;
            break;
        case SDLK_s:
            move.s = !move.s;
            break;
        case SDLK_a:
            move.a = !move.a;
            break;
        case SDLK_d:
            move.d = !move.d;
            break;
        case SDLK_q:
            move.q = !move.q;
            break;
        case SDLK_e:
            move.e = !move.e;
            break;
        default:
            break;
        } 
    }
}


void CameraRoamControl::on_mouse_move( Window* window, int x, int y )
{
    int width, height;
    window->get_dimension(&width, &height);

    Vector2 new_cursor;
    new_cursor.x = ((float)x / (float)width - 0.5f) * 2.0f;
    new_cursor.y = ((float)(height-y) / (float)height - 0.5f) * 2.0f;
 
    Vector2 diff = new_cursor - cursor;

    if( SDL_GetMouseState(NULL, NULL) == SDL_BUTTON(1) )
    {
        Vector3 up = camera.get_up();
        Vector3 side = up.cross( camera.get_direction() );
        camera.rotate(up, -diff.x);
        camera.rotate(side, -diff.y);
    }

    if( SDL_GetMouseState(NULL, NULL) == SDL_BUTTON(3) )
    {
        Vector3 dir = camera.get_direction();
        camera.rotate(-dir, -diff.y);
    }

    cursor = new_cursor;
}


void CameraRoamControl::update( real_t dt )
{
     // control camera movement
    double distance = move.speed * dt;
    Vector3 dir = camera.get_direction().normalize();
    Vector3 up = camera.get_up().normalize();
    Vector3 left = up.cross(dir).normalize();
    Vector3 right = dir.cross(up).normalize();

    if(move.w)
        camera.free_translate(dir * distance);
    if(move.s)
        camera.free_translate(-dir * distance);
    if(move.a)
        camera.free_translate(left * distance);
    if(move.d)
        camera.free_translate(right * distance);
    if(move.q)
        camera.free_translate(-up * distance);
    if(move.e)
        camera.free_translate(up * distance);

}

}
