/**
 * @file color.hpp
 * @brief Color3 class.
 *
 * @author Eric Butler (edbutler)
 */
 
#ifndef _462_MATH_COLOR_HPP_
#define _462_MATH_COLOR_HPP_

#include "math/math.hpp"
#include <iostream>
#include <cassert>

namespace _462 {

/**
 * Represents an RGB color.
 */
class Color3
{
public:

    /**
     * The number of dimensions.
     */
    static const size_t DIM = 3;

    static const Color3 Black;
    static const Color3 White;
    static const Color3 Red;
    static const Color3 Green;
    static const Color3 Blue;

    /**
     * Components of this color.
     */
    real_t r, g, b;

    /**
     * Default constructor. Leaves values unitialized.
     */
    Color3() {}

    /**
     * Create a color with the given values.
     */
    Color3( real_t r, real_t g, real_t b )
		: r( r ), g( g ), b( b ) {}

    /**
     * Create a color from the elements of the given array.
     */
    explicit Color3( const float p[DIM] )
		: r( p[0] ), g( p[1] ), b( p[2] ) {}

    /**
     * Create a color from the elements of the given array.
     */
    explicit Color3( const double p[DIM] )
		: r( p[0] ), g( p[1] ), b( p[2] ) {}

    // also uses default copy and assignment

    Color3 operator*( const Color3& rhs ) const {
        return Color3( r * rhs.r, g * rhs.g, b * rhs.b );
    }

    Color3& operator*=( const Color3& rhs ) {
        r *= rhs.r;
        g *= rhs.g;
        b *= rhs.b;
        return *this;
    }

    /**
     * @remark No bounds checking.
     */
    real_t operator[]( size_t i ) const {
		// assumes all members are in a contiguous block
        assert( i >= 0 && i < DIM );
        return ( &r )[i];
    }

    /**
     * @remark No bounds checking.
     */
    real_t& operator[]( size_t i ) {
		// assumes all members are in a contiguous block
        assert( i >= 0 && i < DIM );
        return ( &r )[i];
    }

    void clamp_zero_one()
    {
        r = std::min( 1.0, std::max( 0.0, r ) );
        g = std::min( 1.0, std::max( 0.0, g ) );
        b = std::min( 1.0, std::max( 0.0, b ) );
    }

    void to_array( float arr[DIM] ) const {
        arr[0] = float( r );
        arr[1] = float( g );
        arr[2] = float( b );
    }

    void to_array( double arr[DIM] ) const {
        arr[0] = double( r );
        arr[1] = double( g );
        arr[2] = double( b );
    }
};

/**
 * Outputs a color text formatted as "(r,g,b)".
 */
std::ostream& operator<<( std::ostream& os, const Color3& rhs );

} /* _462 */

#endif /* _462_MATH_COLOR_HPP_ */

