/**
 * @file model.cpp
 * @brief Mesh class
 *
 * @author Eric Butler (edbutler)
 *           Zeyang Li (zeyangl)
 */

#include "scene/mesh.hpp"
#include "application/opengl.hpp"
#include <iostream>
#include <cstring>
#include <string>
#include <fstream>
#include <sstream>

namespace _462 {

struct Face
{
    size_t num_vertex;
    unsigned int vi[4];
    unsigned int ni[4];
    unsigned int uvi[4];
};

enum ObjFormat
{
    VERTEX_ONLY = 1 << 0,
    VERTEX_UV = 1 << 1,
    VERTEX_NORMAL = 1 << 2,
    VERTEX_UV_NORMAL = 1 << 3
};

Mesh::Mesh() { }
Mesh::~Mesh() { }

bool Mesh::load_from_obj()
{
    std::cout << "Loading model from " << filename << std::endl;

    std::string line;
    std::ifstream file( filename.c_str() );

    typedef std::vector< Vector3 > PositionList;
    typedef std::vector< Vector3 > NormalList;
    typedef std::vector< Vector2 > UVList;
    typedef std::vector< Face > FaceList;

    static const char* scan_vertex = "%d";
    static const char* scan_vertex_uv = "%d/%d";
    static const char* scan_vertex_normal = "%d//%d";
    static const char* scan_vertex_uv_normal = "%d/%d/%d";

    FaceList face_list;
    PositionList position_list;
    NormalList normal_list;
    UVList uv_list;

    int line_num = 0;

    std::string token;

    triangles.clear();

    ObjFormat format = VERTEX_ONLY;

    if ( !file.is_open() ) {
        std::cout << "Error opening file " << filename << " for model loading.\n";
        return false;
    }

    while ( getline( file, line ) )
    {
        std::stringstream stream( line );
        stream >> token;
        line_num++;

        if ( token == "v" ) {

            Vector3 position;
            stream >> position.x >> position.y >> position.z;

            if ( stream.fail() ) {
                std::cerr << "position syntax error on line " << line_num << std::endl;
                return false;
            }

            position_list.push_back( position );

        } else if ( token == "vn" ) {
            Vector3 normal;
            stream >> normal.x >> normal.y >> normal.z;

            if( stream.fail() ) {
                std::cerr << "normal syntax error on line " << line_num << std::endl;
                return false;
            }
            normal_list.push_back( normal );
 
        } else if ( token == "vt" ) {

            Vector2 uv;
            stream >> uv.x >> uv.y;

            if ( stream.fail() ) {
                std::cerr << "uv syntax error on line " << line_num << std::endl;
                return false;
            }

            uv_list.push_back( uv );

        } else if ( token == "f" ) {

            Face face;

            std::vector< std::string > face_tokens;
            std::string vert;

            while ( true ) {
                stream >> vert;
                if( stream.fail() )
                    break;
                face_tokens.push_back( vert );
            }

            // if it's the first time parsing a face, figure out the face format
             if ( face_list.size() == 0 ) {
                std::string token = face_tokens[0];

                if ( token.find( "//" ) != std::string::npos ) {
                    format = VERTEX_NORMAL;
                    std::cout << "loading vertex//normal" << std::endl;
                   } else if ( token.find( '/' ) == std::string::npos ) {
                    format = VERTEX_ONLY;
                    std::cout << "loading vertex only" << std::endl;
                } else {
                    size_t p1 = token.find( '/' );
                    size_t p2 = token.rfind( '/' );
                    if ( p1 == p2 ) {
                        format = VERTEX_UV;
                        std::cout << "loading vertex/uv" << std::endl;
                    } else {
                        format = VERTEX_UV_NORMAL;
                        std::cout << "loading vertex/uv/normal" << std::endl;
                    }
                }
            }


            face.num_vertex = face_tokens.size();

            if ( face.num_vertex > 4 || face.num_vertex < 3 ) {
                std::cerr << "Syntax error at line " << line_num
                          << ", face has incorrect number of vertices" << std::endl;
                return false;
            }

            for ( size_t i = 0; i < face.num_vertex; ++i ) {
                switch ( format )
                {
                    case VERTEX_ONLY:
                        sscanf( face_tokens[i].c_str(),
                                scan_vertex,
                                &face.vi[i] );
                        break;

                    case VERTEX_UV:
                        sscanf( face_tokens[i].c_str(),
                                scan_vertex_uv,
                                &face.vi[i],
                                &face.uvi[i] );
                        break;

                    case VERTEX_NORMAL:
                        sscanf( face_tokens[i].c_str(),
                                scan_vertex_normal,
                                &face.vi[i],
                                &face.ni[i] );
                        break;

                    case VERTEX_UV_NORMAL:
                        sscanf( face_tokens[i].c_str(),
                                scan_vertex_uv_normal,
                                &face.vi[i],
                                &face.uvi[i],
                                &face.ni[i] );
                        break;

                    default:
                        std::cerr << "Syntax error, unrecongnized face format at line "
                                  << line_num << std::endl;
                        break;
                }
            }

            face_list.push_back( face );

        } else if ( token == " " ) {

        } else {
            //std::cerr << "Unknown token on line " << line_num << std::endl;
        }

        token.clear();
        line.clear();
    }

    // construct vertex array from position data
     for ( size_t i = 0; i < position_list.size(); ++i ) {
        Vertex vertex;
        vertex.position = position_list[i];
        vertices.push_back( vertex );
    }

    // fill uvs into vertex array  according to face data
    for ( size_t i = 0; i < face_list.size(); ++i ) {
        Face& face = face_list[i];
          // Quad
        if ( face.num_vertex == 4 ) {
            Triangle tri1 = { { face.vi[0] - 1, face.vi[1] - 1, face.vi[3] - 1 } };
            Triangle tri2 = { { face.vi[1] - 1, face.vi[2] - 1, face.vi[3] - 1 } };
            triangles.push_back( tri1 );
            triangles.push_back( tri2 );
        }
        // Triangle
        else if ( face.num_vertex == 3 ) {
            Triangle tri =  { {face.vi[0] - 1, face.vi[1] - 1, face.vi[2] - 1 } };
            triangles.push_back( tri );
        }

        if ( format == VERTEX_UV_NORMAL || format == VERTEX_UV ) {
            for ( size_t j = 0; j < face.num_vertex; ++j ) {
                unsigned int vi = face.vi[j] - 1;
                unsigned int uvi = face.uvi[j] - 1;
                vertices[vi].tex_coord = uv_list[uvi];
            }
        }

         if ( format == VERTEX_UV_NORMAL || format == VERTEX_NORMAL ) {
            for ( size_t j = 0; j < face.num_vertex; ++j ) {
                unsigned int vi = face.vi[j] - 1;
                unsigned int ni = face.ni[j] - 1;
                vertices[vi].normal = normal_list[ni];
            }
        }
 
    }

    return true;
}

// number of floats per vertex
#define VERTEX_SIZE 8

bool Mesh::load_mesh( bool load_gldata )
{
    // load the model
    if ( !load_from_obj() ) {
        return false;
    }

    if ( !load_gldata ) {
        return true;
    }

    // if no vertices, nothing to do
    if ( vertices.empty() || triangles.empty() ) {
        return false;
    }

    // build vertex data
    vertex_data.resize( vertices.size() * VERTEX_SIZE );
    float* vertex = &vertex_data[0];
    for ( size_t i = 0; i < vertices.size(); ++i ) {
        vertices[i].tex_coord.to_array( vertex + 0 );
        vertices[i].normal.to_array( vertex + 2 );
        vertices[i].position.to_array( vertex + 5 );
        vertex += VERTEX_SIZE;
    }

    // build index data
    index_data.resize( triangles.size() * 3 );
    unsigned int* index = &index_data[0];

    for ( size_t i = 0; i < triangles.size(); ++i ) {
        index[0] = triangles[i].vertices[0];
        index[1] = triangles[i].vertices[1];
        index[2] = triangles[i].vertices[2];
        index += 3;
    }

    return true;
}

void Mesh::render() const
{
    assert( index_data.size() > 0 );
    glInterleavedArrays( GL_T2F_N3F_V3F, VERTEX_SIZE * sizeof vertex_data[0], &vertex_data[0] );
    glDrawElements( GL_TRIANGLES, index_data.size(), GL_UNSIGNED_INT, &index_data[0] );
}

} /* _462 */
