/**
 * @file scene.hpp
 * @brief Class definitions for scenes.
 *
 * @author Eric Butler (edbutler)
 * @author Kristin Siu (kasiu)
 */

#ifndef _462_SCENE_GEOMETRY_HPP_
#define _462_SCENE_GEOMETRY_HPP_

#include "math/vector.hpp"
#include "math/quaternion.hpp"
#include "scene/camera.hpp"
#include "scene/material.hpp"
#include "scene/mesh.hpp"
#include <string>
#include <vector>

namespace _462 {


class Geometry
{
public:
    Geometry();
    virtual ~Geometry();

    /*
       World transformation are applied in the following order:
       1. Scale
       2. Orientation
       3. Position
    */

    // The world position of the object.
    Vector3 position;
    // The world orientation of the object.
    Quaternion orientation;
    // The world scale of the object.
    Vector3 scale;
    // The material for this geometry
    const Material* material;

    /**
     * Renders this geometry using OpenGL in the local coordinate space.
     */
    virtual void render() const = 0;
};


struct PointLight
{
    struct Attenuation
    {
        real_t constant;
        real_t linear;
        real_t quadratic;
    };

    PointLight();

    // The position of the light, relative to world origin.
    Vector3 position;
    // The color of the light (both diffuse and specular)
    Color3 color;
    // attenuation
    Attenuation attenuation;
};

/**
 * The container class for information used to render a scene composed of
 * Geometries.
 */
class Scene
{
public:

    typedef std::vector< PointLight > PointLightList;
    typedef std::vector< Material > MaterialList;
    typedef std::vector< Mesh > MeshList;
    typedef std::vector< Geometry* > GeometryList;

    // the camera
    Camera camera;

    // the background color
    Color3 background_color;
    // the amibient light of the scene
    Color3 ambient_light;
    // the refraction index of air
    real_t refractive_index;

    // list of all lights in the scene
    PointLightList point_lights;
    // all materials used by geometries
    MaterialList materials;
    // all meshes used by models
    MeshList meshes;
    // list of all geometries. deleted in dctor, so should be allocated on heap.
    GeometryList geometries;

    /**
     * Creates a new empty scene.
     */
    Scene();

    /**
     * Destroys this scene. Invokes delete on everything in geometries.
     */
    ~Scene();

    /**
     * Clears the scene, and invokes delete on everything in geometries.
     */
    void reset();

private:

    // no meaningful assignment or copy
    Scene(const Scene&);
    Scene& operator=(const Scene&);

};

} /* _462 */

#endif /* _462_SCENE_GEOMETRY_HPP_ */

