/**
 * @file sphere.cpp
 * @brief Function defnitions for the Sphere class.
 *
 * @author Kristin Siu (kasiu)
 * @author Eric Butler (edbutler)
 */

#include "scene/sphere.hpp"
#include "application/opengl.hpp"

namespace _462 {

#define SPHERE_NUM_LAT 80
#define SPHERE_NUM_LON 100

#define SPHERE_NUM_VERTICES ( ( SPHERE_NUM_LAT + 1 ) * ( SPHERE_NUM_LON + 1 ) )
#define SPHERE_NUM_INDICES ( 6 * SPHERE_NUM_LAT * SPHERE_NUM_LON )

static Vector3 vertices[SPHERE_NUM_VERTICES];
static Vector2 tcoords[SPHERE_NUM_VERTICES];
static unsigned indices[SPHERE_NUM_INDICES];

static void init_sphere()
{
    static bool initialized = false;
    if ( initialized )
        return;

    for ( int i = 0; i <= SPHERE_NUM_LAT; i++ ) {
        for ( int j = 0; j <= SPHERE_NUM_LON; j++ ) {
            real_t lat = real_t( i ) / SPHERE_NUM_LAT;
            real_t lon = real_t( j ) / SPHERE_NUM_LON;

            // texture coordinates are just the longitude and latitude
            tcoords[i * ( SPHERE_NUM_LON + 1 ) + j] = Vector2( lon, 1-lat );

            lat *= PI;
            lon *= 2 * PI;
            real_t sinlat = sin( lat );

            vertices[i * ( SPHERE_NUM_LON + 1 ) + j] =
                Vector3(
                    sinlat * sin( lon ),
                    cos( lat ),
                    sinlat * cos( lon ) );
        }
    }

    for ( int i = 0; i < SPHERE_NUM_LAT; i++ ) {
        for ( int j = 0; j < SPHERE_NUM_LON; j++ ) {
            size_t index = 6 * ( SPHERE_NUM_LON * i + j );
            indices[index + 0] = i * ( SPHERE_NUM_LON + 1 ) + j;
            indices[index + 1] = ( i + 1 ) * ( SPHERE_NUM_LON + 1 ) + j;
            indices[index + 2] = ( i + 1 ) * ( SPHERE_NUM_LON + 1 ) + j + 1;
            indices[index + 3] = indices[index + 0];
            indices[index + 4] = indices[index + 2];
            indices[index + 5] = i * ( SPHERE_NUM_LON + 1 ) + j + 1;
        }
    }

    initialized = true;
}

Sphere::Sphere()
    : radius(0) {}

Sphere::~Sphere() {}

void Sphere::render() const
{
    // create geometry if we haven't already
    init_sphere();

    glEnableClientState( GL_VERTEX_ARRAY );
    glEnableClientState( GL_NORMAL_ARRAY );
    glEnableClientState( GL_TEXTURE_COORD_ARRAY );

    // normals and vertices are the same, if GL_NORMALIZE is enabled
    glVertexPointer(3, GL_DOUBLE, 0, vertices);
    glNormalPointer(GL_DOUBLE, 0, vertices);
    glTexCoordPointer(2, GL_DOUBLE, 0, tcoords);

    glPushMatrix();

    // just scale by radius and draw unit sphere
    glScaled( radius, radius, radius );
    glDrawElements(GL_TRIANGLES, SPHERE_NUM_INDICES, GL_UNSIGNED_INT, indices);

    glPopMatrix();

    glDisableClientState( GL_VERTEX_ARRAY );
    glDisableClientState( GL_NORMAL_ARRAY );
    glDisableClientState( GL_TEXTURE_COORD_ARRAY );
}

} /* _462 */

